/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * Contact Information:
 * Dambach-Werke GmbH
 * Elektronische Leitsysteme
 * Fritz-Minhardt-Str. 1
 * 76456 Kuppenheim
 * Phone: +49-7222-402-0
 * Fax: +49-7222-402-200
 * mailto: info@els.dambach.de
 */

package de.bsvrz.ste.rdstmcvew;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.ClientReceiverInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.ReceiveOptions;
import de.bsvrz.dav.daf.main.ReceiverRole;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.Data.TextValue;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.ConfigurationChangeException;
import de.bsvrz.dav.daf.main.config.MutableSet;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.dambach.parameter.ParameterEmpfaenger;
import de.bsvrz.sys.funclib.dambach.parameter.ParameterSender;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Modelliert eine RDS/TMC-Meldungen so wie sie von der SWE 
 * RDS/TMC-Verwaltung benötigt wird.
 *
 * @author Dambach Werke GmbH
 * @author Stefan Sans
 * @version $Revision: 1.4 $ / $Date: 2008/10/29 12:09:04 $ / ($Author: Hilpp $)
 *
 * Geprüft  Hilpp 30.09.2008
 * Geändert Hilpp 25.10.2008
 */
public class RDSVewMeldung implements ClientReceiverInterface
{
  /**
   * Debug-Logger für Logging-Ausgaben
   */
  private final static Debug _debug = Debug.getLogger();

  /**
   * DAV-Stellevertretet- Objekt vom Typ RdsMeldung
   */
  private SystemObject _davObjekt;

  /**
   * @return das DAV-Stellevertretet- Objekt vom Typ RdsMeldung
   */
  public SystemObject getDavObjekt()
  {
    return _davObjekt;
  }

  /**
   * Übergeordnete Verwaltungs-Instanz
   */
  private RDSVerwaltung _verwaltung;

  /**
   * Datenbeschreibung "versendet"
   */
  public final DataDescription datenBeschreibungVersendet;

  /**
   * Datenbeschreibung "generiert"
   */
  public final DataDescription datenBeschreibungGeneriert;

  /**
   * Datenbeschreibung "bearbeitet"
   */
  public final DataDescription datenBeschreibungBearbeitet;

  /**
   * Ist diese Instanz zum DAV-Empfang angemeldet?
   */
  private boolean _angemeldet = false;

  /**
   * @return ist diese Instanz zum DAV-Empfang angemeldet?
   */
  public boolean isAngemeldet()
  {
    return _angemeldet;
  }

  /**
   * Aktuell empfangene Daten unter dem Aspekt Generiert
   */
  private ResultData _datenGeneriert = null;

  /**
   * Aktuell empfangene Daten unter dem Aspekt Bearbeitet
   */
  private ResultData _datenBearbeitet = null;

  /**
   * Modelliert die Herkunft einer Meldung
   *
   * @author Stefan Sans
   */
  public enum Quelle
  {
    Manuell, Automatisch;

    @Override
    public String toString()
    {
      switch(this)
      {
        case Manuell:
          return "manuell";
        case Automatisch:
          return "automatisch";
      }
      return super.toString();
    }

    /**
     * Bestimmt die Quelle zu einem Text
     * @param text der Text, zu dem die Quelle bestimmt werde soll
     * @return die Quelle zum übergebenen Text
     */
    public static Quelle get(String text)
    {
      for (Quelle quelle: values())
      {
        if (text.equals(quelle.toString()))
        {
          return quelle;
        }
      }
      throw new IllegalArgumentException(text + " ist kein Text zu enum Quelle");
    }
  }

  /**
   * Der Status einer Meldung
   *
   * @author Stefan Sans
   */
  public enum Status
  {
    Unbekannt, Modifikation, Aufhebung, Löschung, Erweiterung;

    /**
     * Bestimmt den Status zu einem Text
     * @param text der Text, zu dem der Status bestimmt werde soll
     * @return den Status zum übergebenen Text
     */
    public static Status get(String text)
    {
      return Status.valueOf(text);
    }
  }

  /**
   * Der Zustand einer Meldung
   *
   * @author Stefan Sans
   */
  public enum Zustand
  {
    NichtQuittiert, Quittiert, Editiert, Verworfen;

    @Override
    public String toString()
    {
      switch(this)
      {
        case NichtQuittiert:
          return "nicht quittiert";
        case Quittiert:
          return "quittiert";
        case Editiert:
          return "editiert";
        case Verworfen:
          return "verworfen";
      }
      return super.toString();
    }

    /**
     * Bestimmt den Zustand zu einem Text
     * @param text der Text, zu dem der Zustand bestimmt werde soll
     * @return den Zustand zum übergebenen Text
     */
    public static Zustand get(String text)
    {
      for (Zustand zustand: values())
      {
        if (text.equals(zustand.toString()))
        {
          return zustand;
        }
      }
      throw new IllegalArgumentException(text + " ist kein Text zu enum Zustand");
    }
  }

  /**
   * Bestimmt den Wert des Attributs Quelle eines Datensatzes der atg.rdsMeldung
   *
   * @param data der Datensatz
   * @return der Wert des Attributs
   */
  protected static Quelle getQuelle(Data data)
  {
    Quelle result = null;

    try
    {
      Data vewInfo = data.getItem("Version").getItem("VerwaltungsInformationen");
      TextValue tv = vewInfo.getTextValue("Quelle");
      return Quelle.get(tv.getText());
    }
    catch (RuntimeException e)
    {
      _debug.warning("Quelle kann nicht bestimmt werden für Datum " +
          data.getName(), e);
    }

    return result;
  }

  /**
   * Bestimmt den Wert des Attributs Zustand eines Datensatzes der atg.rdsMeldung
   *
   * @param data der Datensatz
   * @return der Wert des Attributs
   */
  protected static Zustand getZustand(Data data)
  {
    Zustand result = null;

    try
    {
      Data vewInfo = data.getItem("Version").getItem("VerwaltungsInformationen");
      TextValue tv = vewInfo.getTextValue("Zustand");
      return Zustand.get(tv.getText());
    }
    catch (RuntimeException e)
    {
      _debug.warning("Zustand kann nicht bestimmt werden für Datum " +
          data.getName(), e);
    }

    return result;
  }

  /**
   * Bestimmt den Wert des Attributs Status eines Datensatzes der atg.rdsMeldung
   *
   * @param data der Datensatz
   * @return der Wert des Attributs
   */
  protected static Status getStatus(Data data)
  {
    Status result = null;

    try
    {
      Data vewInfo = data.getItem("Version").getItem("VerwaltungsInformationen");
      TextValue tv = vewInfo.getTextValue("Status");
      return Status.get(tv.getText());
    }
    catch (RuntimeException e)
    {
      _debug.warning("Status kann nicht bestimmt werden für Datum " +
          data.getName(), e);
    }

    return result;
  }

 /**
  * Zugeordnetes Objekt vom Typ RDSVewSender - über dieses wird die Meldungspublikation
  * für diese Meldung abgewickelt
  */
  private RDSVewSender _sender = null;

  /**
   * Konstruiert eine Instanz vom Typ RDSVewMeldung
   *
   * @param davObjekt Systemobjekt vom Typ RdsMeldung
   * @param verwaltung übergeordnetes Verwaltungs-Objekt
   */
  public RDSVewMeldung(SystemObject davObjekt, RDSVerwaltung verwaltung )
  {
    _davObjekt = davObjekt;
    _verwaltung = verwaltung;

    AttributeGroup atg = getConnection().getDataModel().getAttributeGroup("atg.rdsMeldung");
    Aspect asp = getConnection().getDataModel().getAspect("asp.rdsGeneriert");
    datenBeschreibungGeneriert = new DataDescription(atg, asp);

    asp = getConnection().getDataModel().getAspect("asp.rdsBearbeitet");
    datenBeschreibungBearbeitet = new DataDescription(atg, asp);

    asp = getConnection().getDataModel().getAspect("asp.rdsVersendet");
    datenBeschreibungVersendet = new DataDescription(atg, asp);

    _sender = new RDSVewSender(this);

    /*
     * Restauriere unter dem Aspekt RDSBearbeitet empfangene Daten aus den Parametern -
     * da diese Daten als Senke empfangen wurden
     */
    _datenBearbeitet  = new ParameterEmpfaenger(getConnection(), getDavObjekt(),
        "atg.rdsMeldungBearbeitet", "atg.rdsMeldung").getData();

    getConnection().subscribeReceiver(this, davObjekt, datenBeschreibungBearbeitet,
        ReceiveOptions.normal(), ReceiverRole.drain());

    _debug.fine(datenBeschreibungBearbeitet + " zum Empfang angemeldet für " + davObjekt);

    getConnection().subscribeReceiver(this, davObjekt, datenBeschreibungGeneriert,
        ReceiveOptions.normal(), ReceiverRole.receiver());
    _debug.fine(datenBeschreibungGeneriert + " zum Empfang angemeldet für " + davObjekt);

    getConnection().subscribeReceiver(this, davObjekt, datenBeschreibungVersendet,
        ReceiveOptions.normal(), ReceiverRole.receiver());
    _debug.fine(datenBeschreibungVersendet + " zum Empfang angemeldet für " + davObjekt);

    _angemeldet = true;
  }

  /**
   * Gibt die mit dieser RDS/TMC-Meldung verbundenen Ressourcen frei
   */
  public void dispose()
  {
    if (isAngemeldet())
    {
      getConnection().unsubscribeReceiver(this,_davObjekt, datenBeschreibungBearbeitet);
      _debug.fine("Empfang abgemeldet von " + datenBeschreibungBearbeitet + " für " + _davObjekt);

      getConnection().unsubscribeReceiver(this,_davObjekt, datenBeschreibungGeneriert);
      _debug.fine("Empfang abgemeldet von " + datenBeschreibungGeneriert + " für " + _davObjekt);

      getConnection().unsubscribeReceiver(this,_davObjekt, datenBeschreibungVersendet);
      _debug.fine("Empfang abgemeldet von " + datenBeschreibungVersendet + " für " + _davObjekt);

      _angemeldet = false;
    }
    if (_sender != null)
    {
      _sender.dispose();
      _sender = null;
    }
  }

  /**
   * @return die aktuelle DAV-Verbindung
   */
  public ClientDavInterface getConnection()
  {
    return _verwaltung.getConnection();
  }

  /**
   * Setzt den Wert des Attributs Status in einem Datensatz der atg.rdsMeldung
   *
   * @param data Datensatz der atg.rdsMeldung
   * @param status Wert des Attributs Status
   */
  protected static void setStatus(Data data, Status status)
  {
      Data vewInfo = data.getItem("Version").getItem("VerwaltungsInformationen");
      TextValue tv = vewInfo.getTextValue("Status");
      tv.setText(status.toString());
  }

  /**
   * Löscht das aktuelle dynamische Meldungsobjekt - auch die dynamische Menge wird
   * bereinigt.
   * Nach dem Löschen sollte diese Instanz nicht mehr verwendet und dem gc überlassen
   * werden.
   */
  protected void loesche()
  {
    MutableSet menge = _verwaltung.getKonfigObjekt().getMutableSet(RDSVerwaltung.RDSMeldungen);
    if (menge != null)
    {
      try
      {
        menge.remove(getDavObjekt());
      }
      catch (ConfigurationChangeException e)
      {
        throw new RuntimeException("Objekt " + getDavObjekt() + " konnte nich aus Menge " +
            menge + "entfernt werden", e);
      }

      try
      {
        getDavObjekt().invalidate();
      }
      catch (ConfigurationChangeException e)
      {
        throw new RuntimeException("Objekt " + getDavObjekt() + " konnte nich gelöscht werden", e);
      }
    }
    else
    {
      throw new IllegalStateException("Dynamische Menge \"" +
          RDSVerwaltung.RDSMeldungen + "\" nicht gefunden für "  + _verwaltung.getKonfigObjekt());
    }
  }

  /*
   * (Kein Javadoc)
   * @see de.bsvrz.dav.daf.main.ClientReceiverInterface#update(de.bsvrz.dav.daf.main.ResultData[])
   */
  public void update(ResultData[] resultData)
  {
    for (ResultData data : resultData)
    {
      Aspect asp = data.getDataDescription().getAspect();
      AttributeGroup atg = data.getDataDescription().getAttributeGroup();

      if (atg.equals(datenBeschreibungGeneriert.getAttributeGroup()) &&
          asp.equals(datenBeschreibungGeneriert.getAspect()))
      {
        if (_angemeldet)
        {
          _debug.fine("Update für " + data.getObject() + " mit " + data.getDataDescription());
          _datenGeneriert = data;
          bearbeiteDaten(data);
        }
        else
        {
          _debug.warning("Update nicht erwartet, da Datenbeschreibung für dieses Objekt bereits abgemeldet oder nicht angemeldet: "
              + "\tAktuelle Zeit    : " + getConnection().getTime()
              + "\tZeitstempel      : " + data.getDataTime()
              + "\tObjekt           : "+ data.getObject()
              + "\tDatenbeschreibung: " + data.getDataDescription()
              );
        }
      }

      if (atg.equals(datenBeschreibungVersendet.getAttributeGroup()) &&
          asp.equals(datenBeschreibungVersendet.getAspect()))
      {
        if (_angemeldet)
        {
          _debug.fine("Update für " + data.getObject() + " mit " + data.getDataDescription());

  //        System.out.println("RDSVewMeldung:  ---- Update für " + data.getObject() + " mit " + data.getDataDescription());

          if (data.hasData())
          {
  	        Data datenSatz = data.getData().createModifiableCopy();
  	        long zeitStempel = data.getDataTime();

  	        Quelle quelle = getQuelle(datenSatz);
  	        Zustand zustand = getZustand(datenSatz);
  	        Status status = getStatus(datenSatz);

  //	        System.out.println("Bearbeite empfangene Daten für " + getDavObjekt().getPid() +
  //	        " Aspekt: " + asp.getName() +  " Quelle: " + quelle + " Zustand: " + zustand +
  //	        " Status: " + status + " ZeitStempel: " + zeitStempel);
  //
  	        if( status.name() == "Löschung"  )  // Lösche Objekt
  	        {
  	          loesche();
  	        }
  	      }
        }
        else
        {
          _debug.warning("Update nicht erwartet, da Datenbeschreibung für dieses Objekt bereits abgemeldet oder nicht angemeldet: "
              + "\tAktuelle Zeit    : " + getConnection().getTime()
              + "\tZeitstempel      : " + data.getDataTime()
              + "\tObjekt           : "+ data.getObject()
              + "\tDatenbeschreibung: " + data.getDataDescription()
              );
        }
      }

      if (atg.equals(datenBeschreibungBearbeitet.getAttributeGroup()) &&
          asp.equals(datenBeschreibungBearbeitet.getAspect()))
      {
        if (_angemeldet)
        {
          _debug.fine("Update für " + data.getObject() + " mit " + data.getDataDescription());
          _datenBearbeitet = data;
          bearbeiteDaten(data);
          if (data.hasData())
          {
            /*
             * Persistiere unter dem Aspekt RDSBearbeitet empfangene Daten als Parameter -
             * da diese Daten als Senke empfangen wurden
             */
            new ParameterSender(getConnection(), getDavObjekt(),
                data, "atg.rdsMeldungBearbeitet").start();
          }
        }
        else
        {
          _debug.warning("Update nicht erwartet, da Datenbeschreibung für dieses Objekt bereits abgemeldet oder nicht angemeldet: "
              + "\tAktuelle Zeit    : " + getConnection().getTime()
              + "\tZeitstempel      : " + data.getDataTime()
              + "\tObjekt           : "+ data.getObject()
              + "\tDatenbeschreibung: " + data.getDataDescription()
              );
        }
      }
    }
  }

  /**
   * Prüft die übergebenen Daten und gibt sie ggf. zum Senden frei oder loggt
   * entsprechende Fehlermeldung ins
   *
   * @param data die aktuell weiterzuleitenden Daten
   */
  private void bearbeiteDaten(ResultData data)
  {
    if (!data.hasData())
    {
      return;
    }

    Aspect asp = data.getDataDescription().getAspect();
    Data datenSatz = data.getData().createModifiableCopy();
    long zeitStempel = data.getDataTime();

    Quelle quelle = getQuelle(datenSatz);
    Zustand zustand = getZustand(datenSatz);
    Status status = getStatus(datenSatz);

    String text = "Bearbeite empfangene Daten für " + getDavObjekt().getPid() +
    "\nAspekt: " + asp.getName() +  "\nQuelle: " + quelle + "\nZustand: " + zustand +
    "\nStatus: " + status + "\nZeitStempel: " + zeitStempel;
    String suffixFehler = "\nEmpfangene Daten werden NICHT weitergeleitet";

    /*
     * Prüfe Plausibilität Quelle/Zustand
     */
    if (asp.equals(datenBeschreibungGeneriert.getAspect()))
    {
      if (quelle == Quelle.Automatisch)
      {
        /*
         * An dieser Stelle sind Daten mit dem Zustand "nicht quittiert" zulässig.
         * Sie werden verzögert weitergeleitet.
         */
        if (zustand != Zustand.NichtQuittiert)
        {
          _debug.error(text + "\nFEHLER: Zustand " + zustand + " - für Quelle " + quelle +
              "/" + asp.getPid() + " ist nur " + Zustand.NichtQuittiert + " zulässig" + suffixFehler);
          return;
        }
      }
      else
      {
        _debug.error(text + "\nFEHLER: Quelle " + quelle + " - für Aspekt " + asp.getName() +
            " ist nur " + Quelle.Automatisch + " zulässig" + suffixFehler);
        return;
      }
    }
    else if (asp.equals(datenBeschreibungBearbeitet.getAspect()))
    {
      if (quelle == Quelle.Automatisch)
      {
        /*
         * An dieser Stelle sind Daten mit dem Zuständen "quittiert", "editiert" und "verworfen"
         * zulässig. Sie werden sofort weitergeleitet.
         */
        if (zustand != Zustand.Quittiert &&
            zustand != Zustand.Editiert &&
            zustand != Zustand.Verworfen)
        {
          _debug.error(text + "\nFEHLER: Zustand " + zustand + " - für Quelle " + quelle +
              "/" + asp.getPid() + " sind nur " + Zustand.Quittiert + ", " + Zustand.Editiert +
              " und " + Zustand.Verworfen + " zulässig" + suffixFehler);
          return;
        }
      }
      else if (quelle == Quelle.Manuell)
      {
        /*
         * An dieser Stelle sind Daten mit dem Zuständen "editiert" und "verworfen" zulässig.
         * Sie werden sofort weitergeleitet.
         */
        if (zustand != Zustand.Editiert &&
            zustand != Zustand.Verworfen)
        {
          _debug.error(text + "\nFEHLER: Zustand " + zustand + " - für Quelle " + quelle +
              "/" + asp.getPid() + " sind nur " + Zustand.Editiert + " und " +
              Zustand.Verworfen + " zulässig" + suffixFehler);
          return;
        }
      }
      else
      {
        _debug.error(text + "\nFEHLER: Quelle " + quelle + " - für Aspekt " + asp.getName() +
            " sind nur " + Quelle.Automatisch + " und " + Quelle.Manuell + " zulässig" +
            suffixFehler);
        return;
      }
    }

    /*
     * Prüfe Zeitstempel generierte Daten: Diese werden nicht weitergeleitet, wenn es bearbeitete
     * Daten oder gesendete Daten mit höherem Zeitstempel gibt.
     */
    if (asp.equals(datenBeschreibungGeneriert.getAspect()))
    {
      if (_datenBearbeitet != null && _datenBearbeitet.hasData() &&
          _datenBearbeitet.getDataTime() > zeitStempel)
      {
        _debug.warning(text + "\nEs liegen bereits bearbeite Daten mit höherem Zeitstempel " +
            _datenBearbeitet.getDataTime() + " vor" + suffixFehler);
        return;
      }
      else if (_sender.getDatenSenden() != null && _sender.getDatenSenden().hasData() &&
          _sender.getDatenSenden().getDataTime() > zeitStempel)
      {
        _debug.warning(text + "\nEs liegen bereits gesendete Daten mit höherem Zeitstempel " +
            _sender.getDatenSenden().getDataTime() + " vor" + suffixFehler);
        return;
      }
    }

    /*
     * Prüfe Zeitstempel bearbeitete Daten: Diese werden nicht weitergeleitet, wenn der Zustand
     * quittiert ist und es generierte Daten mit neuerem Zeitstempel gibt oder wenn es gesendete
     * Daten mit höherem Zeitstempel gibt.
     */
    if (asp.equals(datenBeschreibungBearbeitet.getAspect()))
    {

      if (zustand == Zustand.Quittiert &&
        _datenGeneriert != null && _datenGeneriert.hasData() &&
        _datenGeneriert.getDataTime() > zeitStempel)
      {
        _debug.warning(text + "\nEs liegen bereits generierte Daten mit höherem Zeitstempel " +
            _datenGeneriert.getDataTime() + " vor" + suffixFehler);
        return;
      }
      else if (_sender.getDatenSenden() != null && _sender.getDatenSenden().hasData() &&
          _sender.getDatenSenden().getDataTime() > zeitStempel)
      {
        _debug.warning(text + "\nEs liegen bereits gesendete Daten mit höherem Zeitstempel " +
            _sender.getDatenSenden().getDataTime() + " vor" + suffixFehler);
        return;
      }
    }

    Data datenSenden = _sender.getDatenSenden() != null && _sender.getDatenSenden().hasData() ?
        _sender.getDatenSenden().getData() : null;

    /*
     * Wenn bereits Status Aufhebung gesendet wurde, wird nur noch Status Löschung weitergeleitet,
     * wenn bereits Status Löschung gesendet wurde, werden keine Daten mehr weitergeleitet.
     */
    if (datenSenden != null)
    {
      Status sendeStatus = getStatus(datenSenden);
      if (sendeStatus == Status.Aufhebung)
      {
        if (status != Status.Löschung || true )
        {
          _debug.config(text + "\nEs liegt bereits eine Aktualisierung mit Status " + sendeStatus +
              " vor - nur Aktualisierungen mit Status " + Status.Löschung + " werden weitergeleitet"
              + suffixFehler);
          return;
        }
      }
      else if (sendeStatus == Status.Löschung)
      {
        if (!(status == Status.Löschung && _verwaltung.getLoeschVerfahren() == 1))
        {
          _debug.config(text + "\nEs liegt bereits eine Aktualisierung mit Status " + sendeStatus +
              " vor" + suffixFehler);
          return;
        }
      }
    }

    /*
     * Zu sendende Daten unter dem Aspekt RDSGeneriert werden dann nicht weitergeleitet, wenn
     * aktuell bearbeitete Daten mit dem Zustand editiert oder verworfen vorliegen
     */
    if (asp.equals(datenBeschreibungGeneriert.getAspect()))
    {
      Data datenBearbeitet = _datenBearbeitet != null && _datenBearbeitet.hasData() ?
          _datenBearbeitet.getData() : null;
      Zustand zustandBearbeited = datenBearbeitet != null ?
          getZustand(datenBearbeitet) : null;
      if ((zustandBearbeited == Zustand.Editiert || zustandBearbeited == Zustand.Verworfen) &&
//          !(sendeStatus == Status.Aufhebung && status == Status.Löschung))
          (datenSenden == null || !(getZustand(datenSenden) == Zustand.Verworfen)))
      {
        _debug.config(text + "\nEs liegen bearbeitete Daten im Zustand " + zustandBearbeited +
            " vor" + suffixFehler);
        return;
      }
    }

    /*
     * Wenn zu sendende Daten unter dem Aspekt RDSBearbeitet den Zustand verworfen
     * haben, wird mindestens der Status Aufhebung erzwungen
     */
    if (asp.equals(datenBeschreibungBearbeitet.getAspect()))
    {
      if (zustand == Zustand.Verworfen)
      {
        if (status != Status.Aufhebung && status != Status.Löschung)
        {
          text += "\nDer Status wird auf \"" + Status.Aufhebung +
          "\" gesetzt, da der Zustand \"" + Zustand.Verworfen +"\" ist";
          try
          {
            setStatus(datenSatz, Status.Aufhebung);
          }
          catch(RuntimeException e)
          {
            _debug.error(text + "\nFehler beim Setzen von Status auf " + Status.Aufhebung +
                suffixFehler, e);
            return;
          }
        }
      }
    }

    /*
     * Ggf. Löschen des dynamischen Meldungsobjekts.
     * Nach dem Löschen/Invalidieren des dynamischen Meldungsobjekts werden
     * für diese Instanz keine Aktualiserungen - auch nicht der Status
     * "Löschung" selbst - weitergeleitet
     */
    boolean loeschen = false;

    switch(_verwaltung.getLoeschVerfahren())
    {
      case 0:
        /*
         * Wenn Status Aufhebung oder Löschung gesendet werden soll und bisher
         * noch keine Daten gesendet wurden, wird die Aktualisierung unterdrückt
         * und die Meldung gelöscht.
         */
        if ((status == Status.Aufhebung || status == Status.Löschung) &&
            (_sender.getDatenSenden() == null || !_sender.getDatenSenden().hasData()))
        {
          text += "\nStatus " + status +
              " wurde empfangen und es wurde bisher noch nichts gesendet";
          loeschen = true;
        }
        break;

      case 1:
        /*
         * Wenn Status Aufhebung gesendet werden soll und bisher noch keine
         * Daten gesendet wurden oder wenn Löschung gesendet werden soll, wird
         *  die Aktualisierung unterdrückt und die Meldung gelöscht.
         */
        if (status == Status.Aufhebung &&
            (_sender.getDatenSenden() == null || !_sender.getDatenSenden().hasData()))
        {
          text += "\nStatus " + status +
              " wurde empfangen und es wurde bisher noch nichts gesendet";
          loeschen = true;
        }
        else if (status == Status.Löschung)
        {
          text += "\nStatus " + status + " wurde empfangen";
          loeschen = true;
        }
        break;
    }

    if (loeschen)
    {
      _debug.config(text + suffixFehler + "\nAktuelle Meldung wird gelöscht");
      try
      {
        loesche();
      }
      catch(RuntimeException e)
      {
        _debug.error("Fehler beim Löschen von " + getDavObjekt(), e);
      }
    }
    else if (asp.equals(datenBeschreibungGeneriert.getAspect()))
    {
      /*
       *  Zu sendende Daten unter dem Aspekt RDSGeneriert werden verzögert gesendet
       *  Die Verzögerungs-/Rückhaltezeit wird (bis auf min. 0) nach unten korrigiert,
       *  wenn der Zeitstempel in der Vergangenheit liegt.
       */
      long aktuelleZeit =  getConnection().getTime();
      long verzoegerung = getRDSMeldungsRueckhaltung() * 1000;

      if (!_verwaltung.getOhneZeitstempelKorrektur() && zeitStempel < aktuelleZeit)
      {
        verzoegerung -= (aktuelleZeit - zeitStempel);
        if (verzoegerung < 0)
        {
          verzoegerung = 0;
        }
      }
      text += "\nEmpfangene generierte Daten werden zur Weiterleitung gemäß RDSMeldungsAktion in " +
        Math.round(verzoegerung / 1000) + "s freigegeben";
      _debug.config(text);
      _sender.starteSenden(datenSatz, zeitStempel, verzoegerung, true, asp);
    }
    else if (asp.equals(datenBeschreibungBearbeitet.getAspect()))
    {
      /*
       *  Zu sendende Daten unter dem Aspekt RDSBearbeitet werden sofort gesendet
       */
      text += "\nEmpfangene bearbeitete Daten werden zur sofortigen Weiterleitung freigegeben";
      _debug.config(text);
      _sender.starteSenden(datenSatz, zeitStempel, 0, true, asp);
    }
  }

  /**
   * @return den Wert des Parameters RDSMeldungsRückhaltung
   */
  public long getRDSMeldungsRueckhaltung()
  {
    return _verwaltung.getRDSMeldungsRueckhaltung();
  }
  
  /**
   * @return den Wert des Parameters RDSMeldungsAktion
   */
  public String getRDSMeldungsAktion()
  {
    return _verwaltung.getRDSMeldungsAktion();
  }
  
}
